/* -*- c-basic-offset: 8 -*-
   rdesktop: A Remote Desktop Protocol client.
   Protocol services - Multipoint Communications Service
   Copyright (C) Matthew Chapman 1999-2002
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "rdesktop.h"

uint16 g_mcs_userid;
extern VCHANNEL g_channels[];
extern unsigned int g_num_channels;

/* Parse an ASN.1 BER header */
static BOOL
ber_parse_header(STREAM s, int tagval, int *length)
{
	int tag, len;

	if (tagval > 0xff)
	{
		in_uint16_be(s, tag);
	}
	else
	{
	in_uint8(s, tag)}

	if (tag != tagval)
	{
		error("expected tag %d, got %d\n", tagval, tag);
		return False;
	}

	in_uint8(s, len);

	if (len & 0x80)
	{
		len &= ~0x80;
		*length = 0;
		while (len--)
			next_be(s, *length);
	}
	else
		*length = len;

	return s_check(s);
}

/* Output an ASN.1 BER header */
static void
ber_out_header(STREAM s, int tagval, int length)
{
	if (tagval > 0xff)
	{
		out_uint16_be(s, tagval);
	}
	else
	{
		out_uint8(s, tagval);
	}

	if (length >= 0x80)
	{
		out_uint8(s, 0x82);
		out_uint16_be(s, length);
	}
	else
		out_uint8(s, length);
}

/* Output an ASN.1 BER integer */
static void
ber_out_integer(STREAM s, int value)
{
	ber_out_header(s, BER_TAG_INTEGER, 2);
	out_uint16_be(s, value);
}

static void
ber_out_uint8(STREAM s, uint8 value)
{
	ber_out_header(s, BER_TAG_INTEGER, 1);
	out_uint8(s, value);
}

static void
ber_in_integer(STREAM s, int *value)
{
	int length;
	ber_parse_header(s, BER_TAG_INTEGER, &length);
	in_uint16_be(s, *value);
}

/* Output a DOMAIN_PARAMS structure (ASN.1 BER) */
static void
mcs_out_domain_params(STREAM s, int max_channels, int max_users, int max_tokens, int max_pdusize)
{
	ber_out_header(s, MCS_TAG_DOMAIN_PARAMS, 26);
	ber_out_uint8(s, 34); // max_channels
	ber_out_uint8(s, 3);  // max_users
	ber_out_uint8(s, 0);  // max_tokens
	ber_out_uint8(s, 1);  // num_priorities 
	ber_out_uint8(s, 0);  // min_throughput
	ber_out_uint8(s, 1);  // max_height
	ber_out_header(s, BER_TAG_INTEGER, 3); // pdu size
	out_uint8(s, 0x00);
	out_uint8(s, 0xff);
	out_uint8(s, 0xf8);
	ber_out_uint8(s, 2);  // ver_protocol
}

/* Parse a DOMAIN_PARAMS structure (ASN.1 BER) */
static BOOL
mcs_parse_domain_params(STREAM s)
{
	int length;
	int max_channels, max_users, max_tokens, max_pdusize;
	int num_priorities, min_throughput, max_height;
	int ver_protocol;

	ber_parse_header(s, MCS_TAG_DOMAIN_PARAMS, &length);
	printf("MCS_TAG_DOMAIN_PARAMS, len %u (expected 32)\n", length);
	if (length == 32) {
		ber_in_integer(s, &max_channels);
		ber_in_integer(s, &max_users);
		ber_in_integer(s, &max_tokens);
		ber_in_integer(s, &num_priorities);
		ber_in_integer(s, &min_throughput);
		ber_in_integer(s, &max_height);
		ber_in_integer(s, &max_pdusize);
		ber_in_integer(s, &ver_protocol);

		printf("max_channels=%u\n", max_channels);
		printf("max_users=%u\n", max_users);
		printf("max_tokens=%u\n", max_tokens);
		printf("num_priorities=%u\n", num_priorities);
		printf("min_throughput=%u\n", min_throughput);
		printf("max_pdusize=%u\n", max_pdusize);
		printf("ver_protocol=%u\n", ver_protocol);
	} else {
		hexdump(s->p, length);
		in_uint8s(s, length);
	}
	
	return s_check(s);
}

/* Expect a MCS_CONNECT_RESPONSE message (ASN.1 BER) */
BOOL
mcs_recv_connect_initial()
{
	uint8 result;
	int length;
	STREAM s;
	char *buf;

	s = iso_recv();
	if (s == NULL)
		return False;

	ber_parse_header(s, MCS_CONNECT_INITIAL, &length);
	printf("parsing MCS_CONNECT_INITIAL (len=%u)\n", length);
	ber_parse_header(s, BER_TAG_OCTET_STRING, &length);   /* calling domain */
	in_uint8(s, result);
	ber_parse_header(s, BER_TAG_OCTET_STRING, &length);   /* called domain */
	in_uint8(s, result);
	
	ber_parse_header(s, BER_TAG_BOOLEAN, &length);
	in_uint8(s, result);

	mcs_parse_domain_params(s);
	mcs_parse_domain_params(s);
	mcs_parse_domain_params(s);

	ber_parse_header(s, BER_TAG_OCTET_STRING, &length);
	in_uint8p(s, buf, length);

	printf("Data from MCS connect: '%*s'\n", length, buf);
	
	return s_check_end(s);
}

/* key generated with "openssl genrsa 512 | openssl rsa -text":
 
modulus:
    00:b1:e0:36:2f:fd:dd:8b:d6:64:e9:2b:14:f8:b9:
    0b:ba:3b:b7:0a:f1:f3:97:56:93:38:01:2f:d1:31:
    2d:70:59:de:97:6a:61:3f:cb:a4:b4:12:05:89:14:
    d9:b0:8a:70:03:b6:f1:ad:c5:b9:19:9b:b9:8f:03:
    51:bf:fe:f8:e5
publicExponent: 65537 (0x10001)
privateExponent:
    00:a4:86:68:58:97:8d:f6:2c:06:06:8d:ac:c6:2a:
    12:a8:dd:56:ff:2e:b0:4b:08:ee:fe:dc:4a:28:4a:
    3e:67:2d:8e:08:6b:5f:87:69:a4:c5:a0:e7:50:1a:
    f0:71:03:46:f2:52:7c:e6:09:40:40:61:51:76:32:
    f8:28:72:99:41
prime1:
    00:db:46:d9:34:10:ce:0d:f6:f7:48:30:0f:2b:f6:
    4d:66:40:27:00:97:db:48:77:cd:96:b5:a6:51:81:
    a1:7c:ad
prime2:
    00:cf:aa:5c:d3:8a:0d:d4:91:a6:40:92:ff:2c:31:
    d6:f9:08:8d:e8:d2:a8:2f:89:a6:88:2c:61:68:8e:
    ee:7c:19
*/
unsigned char my_modulus[SEC_MODULUS_SIZE] = {
	0xe5, 0xf8, 0xfe, 0xbf, 0x51,
	0x03, 0x8f, 0xb9, 0x9b, 0x19, 0xb9, 0xc5, 0xad, 0xf1, 0xb6, 0x03, 0x70, 0x8a, 0xb0, 0xd9,
	0x14, 0x89, 0x05, 0x12, 0xb4, 0xa4, 0xcb, 0x3f, 0x61, 0x6a, 0x97, 0xde, 0x59, 0x70, 0x2d,
	0x31, 0xd1, 0x2f, 0x01, 0x38, 0x93, 0x56, 0x97, 0xf3, 0xf1, 0x0a, 0xb7, 0x3b, 0xba, 0x0b,
	0xb9, 0xf8, 0x14, 0x2b, 0xe9, 0x64, 0xd6, 0x8b, 0xdd, 0xfd, 0x2f, 0x36, 0xe0, 0xb1
};
unsigned char my_exponent[SEC_EXPONENT_SIZE] = {
	0x01, 0x00, 0x01, 0x00
};
unsigned char my_private_exponent[] = {
	0x41, 0x99, 0x72, 0x28, 0xf8,
	0x32, 0x76, 0x51, 0x61, 0x40, 0x40, 0x09, 0xe6, 0x7c, 0x52, 0xf2, 0x46, 0x03, 0x71, 0xf0,
	0x1a, 0x50, 0xe7, 0xa0, 0xc5, 0xa4, 0x69, 0x87, 0x5f, 0x6b, 0x08, 0x8e, 0x2d, 0x67, 0x3e,
	0x4a, 0x28, 0x4a, 0xdc, 0xfe, 0xee, 0x08, 0x4b, 0xb0, 0x2e, 0xff, 0x56, 0xdd, 0xa8, 0x12,
	0x2a, 0xc6, 0xac, 0x8d, 0x06, 0x06, 0x2c, 0xf6, 0x8d, 0x97, 0x58, 0x68, 0x86, 0xa4
};

 
void
mcs_send_connect_response()
{
	STREAM s;
	int i, length;

	s = iso_init(250);
	printf("INITLEN: %u\n", s->p - s->iso_hdr);

	ber_out_header(s, MCS_CONNECT_RESPONSE, 245);
	ber_out_header(s, BER_TAG_RESULT, 1);
	out_uint8(s, 0);

	ber_out_header(s, BER_TAG_INTEGER, 1);
	out_uint8(s, 0);  // connect id
	
	mcs_out_domain_params(s, 34, 2, 0, 0xffff);  // dumdidum?

	ber_out_header(s, BER_TAG_OCTET_STRING, 207);

	// some unknown header of sorts
	out_uint8(s, 0x00);
	out_uint8(s, 0x05);
	out_uint8(s, 0x00);
	out_uint8(s, 0x14);
	out_uint8(s, 0x7c);
	out_uint8(s, 0x00);
	out_uint8(s, 0x01);
	out_uint8(s, 0x2a);
	out_uint8(s, 0x14);
	out_uint8(s, 0x76);
	out_uint8(s, 0x0a); 
	out_uint8(s, 0x01);
	out_uint8(s, 0x01);
	out_uint8(s, 0x00);
	out_uint8(s, 0x01);
	out_uint8(s, 0xc0);
	out_uint8(s, 0x00);
	out_uint8(s, 0x4d);
	out_uint8(s, 0x63);
	out_uint8(s, 0x44);
	out_uint8(s, 0x6e);

	length = 184;
	
	// two bytes of length
	out_uint8(s, 0x80 | (length >> 7));
	out_uint8(s, length & 0x7f);

	// server info -- we claim to support RDP5
	out_uint16_le(s, SEC_TAG_SRV_INFO);
	out_uint16_le(s, 8);  // length
	out_uint16_le(s, 4);  // version
	out_uint16_le(s, 8);  // unknown

	// channel info -- open a few channels
	out_uint16_le(s, SEC_TAG_SRV_CHANNELS);
	out_uint16_le(s, 16); // length
	out_uint16_le(s, 1003);
	out_uint16_le(s, 3);
	out_uint16_le(s, 1004);
	out_uint16_le(s, 1005);
	out_uint16_le(s, 1006);
	out_uint16_le(s, 0);
	
	// crypto info
	out_uint16_le(s, SEC_TAG_SRV_CRYPT);
	out_uint16_le(s, 160); // length
	out_uint32_le(s, 1); // 40-bit
	out_uint32_le(s, 1); // low

	out_uint32_le(s, SEC_RANDOM_SIZE); // random_len
	out_uint32_le(s, 96);             // rsa_info_len
	out_uint8s(s, SEC_RANDOM_SIZE);    // server_random
	out_uint32_le(s, 1);               // RDP4-style
	out_uint8s(s, 8);                  // unknown

	out_uint16_le(s, SEC_TAG_PUBKEY);
	out_uint16_le(s, 88);
	out_uint32_le(s, SEC_RSA_MAGIC);
	out_uint32_le(s, SEC_MODULUS_SIZE + SEC_PADDING_SIZE);  // modulus_len
	out_uint32_le(s, SEC_MODULUS_SIZE * 8);                 // modulus_bits
	out_uint8s(s, 4);                                       // unknown
	out_uint8p(s, my_exponent, SEC_EXPONENT_SIZE);
	out_uint8p(s, my_modulus, SEC_MODULUS_SIZE);
	out_uint8s(s, SEC_PADDING_SIZE);
	
	s_mark_end(s);
	iso_send(s);

}

/* Send an EDrq message (ASN.1 PER) */
static void
mcs_send_edrq(void)
{
	STREAM s;

	s = iso_init(5);

	out_uint8(s, (MCS_EDRQ << 2));
	out_uint16_be(s, 1);	/* height */
	out_uint16_be(s, 1);	/* interval */

	s_mark_end(s);
	iso_send(s);
}

/* Send an AUrq message (ASN.1 PER) */
static void
mcs_send_aurq(void)
{
	STREAM s;

	s = iso_init(1);

	out_uint8(s, (MCS_AURQ << 2));

	s_mark_end(s);
	iso_send(s);
}

/* Send a AUcf message (ASN.1 PER) */
static void
mcs_send_aucf(uint16 mcs_userid)
{
	STREAM s;

	s = iso_init(4);

	out_uint8(s, (MCS_AUCF << 2) | 2);  // | 2 = send user ID
	out_uint8(s, 0);  // success
	out_uint16_be(s, 6);
	
	s_mark_end(s);
	iso_send(s);
}

/* Send a CJrq message (ASN.1 PER) */
static void
mcs_send_cjrq(uint16 chanid)
{
	STREAM s;

	DEBUG_RDP5(("Sending CJRQ for channel #%d\n", chanid));

	s = iso_init(5);

	out_uint8(s, (MCS_CJRQ << 2));
	out_uint16_be(s, g_mcs_userid);
	out_uint16_be(s, chanid);

	s_mark_end(s);
	iso_send(s);
}

/* Expect a CJcf message (ASN.1 PER) */
static void
mcs_send_cjcf(uint16 userid, uint16 chanid)
{
	STREAM s;

	s = iso_init(8);

	out_uint8(s, (MCS_CJCF << 2) | 2);
	out_uint8(s, 0); // success
	out_uint16_be(s, 6);
	out_uint16_be(s, chanid);
	out_uint16_be(s, chanid);

	s_mark_end(s);
	iso_send(s);
}

/* Initialise an MCS transport data packet */
STREAM
mcs_init(int length)
{
	STREAM s;

	s = iso_init(length + 8);
	s_push_layer(s, mcs_hdr, 8);

	return s;
}

/* Send an MCS transport data packet to a specific channel */
void
mcs_send_to_channel(STREAM s, uint16 channel)
{
	uint16 length;

	s_pop_layer(s, mcs_hdr);
	length = s->end - s->p - 8;
	length |= 0x8000;

	out_uint8(s, (MCS_SDIN << 2));
	out_uint16_be(s, g_mcs_userid);
	out_uint16_be(s, channel);
	out_uint8(s, 0x70);	/* flags */
	out_uint16_be(s, length);

	iso_send(s);
}

/* Send an MCS transport data packet to the global channel */
void
mcs_send(STREAM s)
{
	mcs_send_to_channel(s, MCS_GLOBAL_CHANNEL);
}

/* Receive an MCS transport data packet */
STREAM
mcs_recv(uint16 * channel)
{
	uint8 opcode, appid, length, userid;
	STREAM s;

	s = iso_recv();
	if (s == NULL)
		return NULL;

	in_uint8(s, opcode);
	appid = opcode >> 2;

	switch (appid) {
	case MCS_SDRQ:
		in_uint8s(s, 2);	/* userid */
		in_uint16_be(s, *channel);
		in_uint8s(s, 1);	/* flags */
		in_uint8(s, length);
		if (length & 0x80)
			in_uint8s(s, 1);	/* second byte of length */

		return s;
	case MCS_DPUM:
		printf("Received DPUM (?)\n");
		return NULL;
	case MCS_EDRQ:
		// Erect Domain (ignore)
		printf("Received EDrq\n");
		return NULL;
	case MCS_AURQ:
		// Attach User Request, respond with AUcf (Attach User Confirm)
		printf("Received AUrq, sending AUcf\n");
		mcs_send_aucf(0);
		return NULL;
	case MCS_CJRQ:
		// Channel Join Request, respond with CJcf (Channel Join Confirm);
		in_uint16_be(s, userid);
		in_uint16_be(s, *channel);
		printf("Received CJrq for channel %hu, sending CJcf\n", *channel);
		mcs_send_cjcf(userid, *channel);
		return NULL;
	default:
		error("expected data, got %d\n", opcode);
		return NULL;
	}

}

/* Disconnect from the MCS layer */
void
mcs_disconnect(void)
{
	iso_disconnect();
}
